function model_perf_scenario(matrix_filename, filename_with_params_for_matrix, feature_file, algo, tuningScenario, captime_for_matrix)
% MODEL_PERF_SCENARIO: generates diagnostic plots to study how well
% various regression methods predict performance
%
% Input: tuningScenario: also used as prefix for output plots
%        matrix_filename: name of file with I lines for I instances, each
%        line <instance_name>, <performance of config1>, <performance of
%        config2>, ... In my experiments: config1 = default config (first
%        to be explored in IS-best)
%        captime_for_matrix: time after which runs for generating the
%        matrix were terminated -- matrix entries for terminated runs must
%        have values larger than this captime
%        filename_with_params_for_matrix: name of file with C parameter 
%        configurations, where C is the number of columns in the input 
%        matrix, and the results in column i are for configuration i

tuningScenario

%=== Read in a representation of the "function" we optimize, including info
%=== on parameters.
func = read_func(algo, tuningScenario);

%=== Read in matrix data.
data = csvread(matrix_filename,0,1);
matrix = data';
[C,I] = size(matrix); % C: #parameter configurations; I: #instances

%=== Read in configurations used to generate the matrix data.
Theta = read_param_configs(func, filename_with_params_for_matrix, 1);
assert(all(all(Theta >= 1)));

%=== Read in featuers of the instances used to generate the matrix data.
use_features = 1;
[instance_names] = textread(matrix_filename,'%s%*[^\n]', 'bufsize', 100000);
for i=1:length(instance_names)
    instance_names{i} = instance_names{i}(1:end-1);
end
if 0 || ~exist(feature_file, 'file')
    if use_features
        my_feature_file = strcat( matrix_filename, '-features.mat' );
        if exist(my_feature_file, 'file')
            feature_data = load(my_feature_file);
            all_features = feature_data.all_features;
            xNames = feature_data.xNames';
        else
            %=== Startup DB.
            if mysql('status')
                startUpDB
            end
            if union( (strfind(lower(tuningScenario), 'cplex')),  (strfind(lower(tuningScenario), 'gurobi')) )
                mysql('use MODEL_DB');
                func.mip = 1;
            else
                mysql('use MODEL_DB')
                func.mip = 0;
            end

            %=== Read in instance features.
            all_instance_ids = get_inst_id(instance_names(1:I));
            [all_features, xNames] = getFeaturesForID(func.mip, all_instance_ids);
            save(my_feature_file, 'all_features', 'xNames');
        end
        csvwrite(feature_file, [-ones(1,size(all_features,2)+1); ones(size(all_features,1),1), all_features]);
    else
        all_features = zeros(I,1);
    end
else
    all_features = csvread(feature_file, 1, 1);
    xNames = textread(feature_file, '%s', 1, 'whitespace', '\n', 'bufsize', 10000);
    xNames = strread(xNames{1},'%s','whitespace',',');
    xNames = deblank(xNames);
    xNames= xNames(2:end);
end
names = [func.param_names; xNames];

%=== Prepare data using PAR-1.
cutoff_multiplier = 1;
for i=1:size(matrix,2)
    matrix(find(matrix(:,i)>captime_for_matrix-0.01), i) = captime_for_matrix * cutoff_multiplier; 
end
matrix = max(matrix, 0.005);

%=== Shuffle training and test configurations.
perm = randperm(size(matrix,1));
matrix = matrix(perm,:);
Theta = Theta(perm,:);

%=== Shuffle training and test instances.
perm = randperm(size(matrix,2));
matrix = matrix(:,perm);
all_features = all_features(perm,:);


%% After this setup, do some experiments with the matrix.
%% Experiment 1: matrix predictions
if 1
    options_vec = {get_rf_default_options,get_gp_default_options}; %get_gp_default_options};
    seed = 1234;
    matrixPlots = 0;
    for model_idx=1:length(options_vec)
        options = options_vec{model_idx};
        [timeToLearn{model_idx}, timeToPredict{model_idx}, results{model_idx}] = do_matrix_exp(func, matrix, Theta, all_features, names, options, seed, matrixPlots, tuningScenario, 10000, 10000);
    end

    fprintf(strcat([tuningScenario, ' & ']));
    two_output(timeToLearn{1}, timeToLearn{2}, 1);
    fprintf(strcat(['& RMSE']));

    for quadr=1:4
        fprintf(' & ');
        two_output(results{1}{quadr}(1), results{2}{quadr}(1), 1);
    end
    fprintf('\\\\\n');

    fprintf(strcat([' & & & CC']));
    for quadr=1:4
        fprintf(' & ');
        two_output(results{1}{quadr}(3), results{2}{quadr}(3), 0);
    end
    fprintf('\\\\\n');

    fprintf(strcat([' & & & LL']));
    for quadr=1:4
        fprintf(' & ');
        two_output(results{1}{quadr}(2), results{2}{quadr}(2), 0);
    end
    fprintf('\\\\\n');
end

%% Experiment 2: predictions of a single configuration on new instances
if 0
    options_vec = {get_rf_default_options,get_gp_default_options,get_lr_default_options};

    [rmses, lls, ccs, timesToLearn, timesToPredict] = get_inst_pred_perf(options_vec, matrix, Theta, all_features, C, I, 500)
    %=== Plotting of that data.
    rmse_avg(model_idx) = mean(rmses(model_idx,:))
    ll_avg(model_idx) = mean(lls(model_idx,:))
    cc_avg(model_idx) = mean(ccs(model_idx,:))
    boxplot(ccs(1:3,:)')
end

%% Experiment 3: scaling performance with # data points
if 0
    options_vec = {get_rf_default_options,get_gp_default_options};%,get_lr_default_options};

    numTrains = [2,4,8,16,32,64,128,256,512];
    for i=1:length(numTrains)
        numTrain = numTrains(i);
        [rmses, lls, ccs, timesToLearn, timesToPredict] = get_inst_pred_perf(options_vec, matrix, Theta, all_features, C, I, numTrain)
        for model_idx=1:length(options_vec)
            rmse_avg(i,model_idx) = mean(rmses(model_idx,:));
            rmse_std(i,model_idx) = std(rmses(model_idx,:));
            lls_avg(i,model_idx) = mean(lls(model_idx,:));
            lls_std(i,model_idx) = std(lls(model_idx,:));
            cc_avg(i,model_idx) = mean(ccs(model_idx,:));
            cc_std(i,model_idx) = std(ccs(model_idx,:));
            timesToLearn_avg(i,model_idx) = mean(timesToLearn(model_idx,:));
            timesToLearn_std(i,model_idx) = std(timesToLearn(model_idx,:));
            timesToPredict_avg(i,model_idx) = mean(timesToPredict(model_idx,:));
            timesToPredict_std(i,model_idx) = std(timesToPredict(model_idx,:));
        end
    end

    model_names = {'RF', 'PP'};
    plot_Scaling_N(numTrains, cc_avg, cc_std, options_vec, 'Correlation coefficient', model_names, tuningScenario);
    plot_Scaling_N(numTrains, rmse_avg, rmse_std, options_vec, 'RMSE', model_names, tuningScenario, 'NorthEast');
    plot_Scaling_N(numTrains, lls_avg, lls_std, options_vec, 'Test Data Log likelihood', model_names, tuningScenario);
    plot_Scaling_N(numTrains, timesToLearn_avg, timesToLearn_std, options_vec, 'Learn Time', model_names, tuningScenario);
    plot_Scaling_N(numTrains, timesToPredict_avg, timesToPredict_std, options_vec, 'Prediction Time', model_names, tuningScenario);
end





function [rmses, lls, ccs, timesToLearn, timesToPredict] = get_inst_pred_perf(options_vec, matrix, Theta, all_features, C, I, numTrain)
numCruns = min(C,10);
timesToLearn = -ones(length(options_vec), numCruns);
timesToPredict = -ones(length(options_vec), numCruns);
rmses = -ones(length(options_vec), numCruns);
lls = -ones(length(options_vec), numCruns);
ccs = -ones(length(options_vec), numCruns);
perm=randperm(C);
for model_idx=1:length(options_vec)
    fprintf(strcat([' Model ', num2str(model_idx), '/', num2str(length(options_vec)), '...\n']));
    for i=1:numCruns
        fprintf(strcat(['  Configuration ', num2str(i), '/', num2str(numCruns), '...\n']));
        c=perm(i);
        %=== Get indices for training and test instances.
        [theta_idxs, inst_idxs] = get_idx_pairs(c, 1:I);
        [X,y] = getXyForIndices(Theta, all_features, matrix, theta_idxs, inst_idxs);

        %=== Call model_perf with right X and y.
        [rmses(model_idx,i), lls(model_idx,i), ccs(model_idx,i), timesToLearn(model_idx,i), timesToPredict(model_idx,i)] = simple_model_perf(X, y, numTrain, options_vec{model_idx});
    end
end


function func = read_func(algo, tuningScenario)
algo_func_file = strcat(algo, '-func_file.mat');
if exist(algo_func_file, 'file')
    data = load(algo_func_file);
    func = data.func;   
else
    func = readFuncSpecifics(tuningScenario, 0);
    save(algo_func_file, 'func');
end