package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import static ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics.*;

import java.math.BigInteger;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import net.sf.json.JSONObject;

import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;

import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.parameters.BooleanDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.IntegerDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ObjectClassDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpace;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpaceBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.RealDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.analysis.Plot;
import ca.ubc.cs.beta.hal.analysis.Statistics;
import ca.ubc.cs.beta.hal.analysis.Statistics.Alternative;
import ca.ubc.cs.beta.hal.analysis.TestResult;
import ca.ubc.cs.beta.hal.environments.SubrunRunner;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.MetaProblemInstance;
import ca.ubc.cs.beta.hal.utils.Misc;

public class ComprehensivePairwiseComparison extends KWayComparison {
	
	private static final Logger log = Logger.getLogger(ComprehensivePairwiseComparison.class.getCanonicalName());
	
	private static final String VERSION = "1.1.0";
	private static final ParameterSpace DFLT_OUT_SPACE, DFLT_SCN_SPACE, DFLT_CFG_SPACE, SUPP_OPT_SPACE;
	static {
		ParameterSpaceBuilder psb = new ParameterSpaceBuilder();
		psb.put(SCD_OVERLAY, new ObjectClassDomain(Plot.class));
		psb.put(SCATTER, new ObjectClassDomain(Plot.class));
		psb.put(SOLUTION_QUALITY_1, new RealDomain());
		psb.put(MEAN_1, new RealDomain());
		psb.put(STDDEV_1, new RealDomain());
		psb.put(Q90_1, new RealDomain());
		psb.put(Q75_1, new RealDomain());
		psb.put(Q50_1, new RealDomain());
		psb.put(Q25_1, new RealDomain());
		psb.put(Q10_1, new RealDomain());
		psb.put(SCD_1, new ObjectClassDomain(Plot.class));
		psb.put(SOLUTION_QUALITY_2, new RealDomain());
		psb.put(MEAN_2, new RealDomain());
		psb.put(STDDEV_2, new RealDomain());
		psb.put(Q90_2, new RealDomain());
		psb.put(Q75_2, new RealDomain());
		psb.put(Q50_2, new RealDomain());
		psb.put(Q25_2, new RealDomain());
		psb.put(Q10_2, new RealDomain());
		psb.put(SCD_2, new ObjectClassDomain(Plot.class));
		psb.put(SPEARMAN_RHO, new RealDomain(-1., 1.));
		psb.put(SPEARMAN_P, new RealDomain(0., 1.));
		psb.put(WILCOXON_W, new IntegerDomain(0, null));
		psb.put(WILCOXON_P, new RealDomain(0., 1.));
		psb.put(PERMUTATION_P, new RealDomain(0., 1.));
		psb.put(Semantics.RUNLENGTH, Semantics.getDomain(Semantics.RUNLENGTH));
		psb.put(WILCOXON_WINNER, new ObjectClassDomain(ParameterlessAlgorithm.class));
		DFLT_OUT_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		DFLT_CFG_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.SEED, Semantics.getDomain(Semantics.SEED));
		psb.put(Semantics.MAX_CPUTIME, new RealDomain(0.0, null, 1e300));
		psb.put(Semantics.MAX_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(MIN_OUTPUT_INTERVAL, new RealDomain(0., null, 60.0));
		psb.put(COMBINE_IDENTICAL_INSTANCES, new BooleanDomain(false));
		psb.put(MAX_PARALLEL_RUNS, new IntegerDomain(1, null, 100));
		DFLT_SCN_SPACE = psb.build();
		
		psb = new ParameterSpaceBuilder();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, new RealDomain(0.0, null, 60.0));
		psb.put(Semantics.MAX_SUBRUN_RUNLENGTH, new IntegerDomain(1L, null, (long)1e18));
		psb.put(Semantics.RUNS_PER_INSTANCE, new IntegerDomain(1L, null, 1L));
		SUPP_OPT_SPACE = psb.build();
	}
	
	private static final Map<String, String> variables = Misc.asMap(
			SCD_OVERLAY, "solution cost distributions",
			SCATTER, "correlation plot",
			SCD_1, "alg. 1 solution cost distribution",
			SOLUTION_QUALITY_1, "alg. 1 solution quality per provided metric",
			MEAN_1, "alg. 1 mean performance",
			STDDEV_1, "alg. 1 performance standard deviation",
			Q90_1, "alg. 1 90th performance percentile",
			Q75_1, "alg. 1 75th performance percentile",
			Q50_1, "alg. 1 median performance",
			Q25_1, "alg. 1 25th performance percentile",
			Q10_1, "alg. 1 10th performance percentile",
			SCD_2, "alg. 2 solution cost distribution",
			SOLUTION_QUALITY_2, "alg. 2 solution quality per provided metric",
			MEAN_2, "alg. 2 mean performance",
			STDDEV_2, "alg. 2 performance standard deviation",
			Q90_2, "alg. 2 90th performance percentile",
			Q75_2, "alg. 2 75th performance percentile",
			Q50_2, "alg. 2 median performance",
			Q25_2, "alg. 2 25th performance percentile",
			Q10_2, "alg. 2 10th performance percentile",
			MIN_OUTPUT_INTERVAL, "minimum logged output interval (s)",
			COMBINE_IDENTICAL_INSTANCES, "collapse instances with identical hash",
			Semantics.MAX_RUNLENGTH, "maximum number of target algorithm runs",
			MAX_PARALLEL_RUNS, "maximum runs to launch in parallel");
	
	public String describeSemantics(String semantics) {
		String out = variables.get(semantics);
		return out == null ? Semantics.describe(semantics) : out;
	}
	public Collection<String> getDefinedSemantics() {
		return variables.keySet();
	}

	public ComprehensivePairwiseComparison() {
		super(ComprehensivePairwiseComparison.class.getSimpleName(), VERSION,
				DFLT_CFG_SPACE, DFLT_SCN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE,
				Semantics.DETERMINISTIC, false);
	}
	
	@Override
	public String getVersion() { return VERSION; }
	
	@SuppressWarnings("unchecked")
	@Override
	public Set<Set<String>> getRequiredTags() {
		return Misc.asSet(Misc.asSet(MetaProblemInstance.TAG, MetaProblemInstance.getNumAlgsTag(2),
				InstanceMetricMetaProblemInstance.TAG));
	}

	@Override
	public MetaAlgorithmRun getRun(AlgorithmRunRequest request,
			SubrunRunner runner, Statistics stats) {
		
		if (!request.getImplementation().equals(this)) {
			throw new IllegalArgumentException("Request does not correspond to this meta-algorithm implementation");
		}
		return new ComprehensivePairwiseComparisonRun(request, runner, stats);
	}

	public static ComprehensivePairwiseComparison fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(ComprehensivePairwiseComparison.class, spec);
		String ver = o.optString("version");
		if (!VERSION.equals(ver) && ver != null && ver.length() > 0) {
			throw new IllegalArgumentException("Record corresponds to different version of ComprehensivePairwiseComparison plugin; was:" +
					ver + "; is: " + VERSION );
		}
		return new ComprehensivePairwiseComparison();
	}
	
	private static class ComprehensivePairwiseComparisonRun extends KWayComparisonRun {

		private final ParameterlessAlgorithm algA, algB;
		private final Statistics stats;
		
		public ComprehensivePairwiseComparisonRun(AlgorithmRunRequest request,
				SubrunRunner runner, Statistics stats) {
			super(request, runner, stats);
			if (!targets.size().equals(BigInteger.valueOf(2))) throw new IllegalArgumentException("must have exactly 2 target algorithms");
			algA = targets.get(0);
			algB = targets.get(1);
			this.stats = stats;
			
			if (!(stats.isPermutationTestSupported())) {
				log.warning("Permutation is not be supported.");
			}
		}
		
		@SuppressWarnings("unchecked")
		@Override
		public Map<String, Object> getOutputSnapshot() {
			
			Map<String, Object> out = new HashMap<String, Object>();
			DescriptiveStatistics listA = new DescriptiveStatistics(),
					listB = new DescriptiveStatistics();
			
			synchronized(data) {
				Map<String, Object> coreout = super.getOutputSnapshot();
				out.put(Semantics.RUNLENGTH, coreout.get(Semantics.RUNLENGTH));
				
				Map<Long, List<Double>> datA = data.get(algA), datB = data.get(algB);
				if (datA == null || datB== null) return out; // just barely started, one run in
				for (Map.Entry<Long, List<Double>> pi : datA.entrySet()) {
					if (datB.containsKey(pi.getKey()) && datB.get(pi.getKey()).size() > 0) {
						listA.addValue(metric.aggregate(pi.getValue()));
						listB.addValue(metric.aggregate(datB.get(pi.getKey())));
					}
				}
				Plot p = Plot.makeScatter(datA, datB, metric);
				p.setXlabel(algA.getName());
				p.setYlabel(algB.getName());
				out.put(SCATTER, p);
				
				out.put(SCD_OVERLAY, coreout.get(SCD_OVERLAY));
				
				out.put(SCD_1, ((List<Plot>)coreout.get(SCDS)).get(0));
				out.put(SCD_2, ((List<Plot>)coreout.get(SCDS)).get(1));
				
				out.put(SOLUTION_QUALITY_1, ((List<Double>)coreout.get(SOLUTION_QUALITIES)).get(0));
				out.put(SOLUTION_QUALITY_2, ((List<Double>)coreout.get(SOLUTION_QUALITIES)).get(1));

				out.put(MEAN_1, ((List<Double>)coreout.get(MEANS)).get(0));
				out.put(MEAN_2, ((List<Double>)coreout.get(MEANS)).get(1));
				out.put(STDDEV_1, ((List<Double>)coreout.get(STDDEVS)).get(0));
				out.put(STDDEV_2, ((List<Double>)coreout.get(STDDEVS)).get(1));

				out.put(Q90_1, ((List<Double>)coreout.get(Q90S)).get(0));
				out.put(Q90_2, ((List<Double>)coreout.get(Q90S)).get(1));
				out.put(Q75_1, ((List<Double>)coreout.get(Q75S)).get(0));
				out.put(Q75_2, ((List<Double>)coreout.get(Q75S)).get(1));
				out.put(Q50_1, ((List<Double>)coreout.get(Q50S)).get(0));
				out.put(Q50_2, ((List<Double>)coreout.get(Q50S)).get(1));
				out.put(Q25_1, ((List<Double>)coreout.get(Q25S)).get(0));
				out.put(Q25_2, ((List<Double>)coreout.get(Q25S)).get(1));
				out.put(Q10_1, ((List<Double>)coreout.get(Q10S)).get(0));
				out.put(Q10_2, ((List<Double>)coreout.get(Q10S)).get(1));
			}
			
			double[] valsA = listA.getValues(), valsB = listB.getValues();
			if (valsA.length > 0 && valsB.length > 0) {
				TestResult tr = stats.spearmanTest(valsA, valsB, Alternative.TWO_SIDED);
				extraCpuTime += tr.getExternalComputeTime(); // if the statistics takes some non-java time, account for it.
				out.put(SPEARMAN_RHO, tr.getStatistic());
				out.put(SPEARMAN_P, tr.getP());
				tr = stats.wilcoxonTest(valsA, valsB, 0.0, Alternative.TWO_SIDED, true);
				extraCpuTime += tr.getExternalComputeTime(); // if the statistics takes some non-java time, account for it.
				out.put(WILCOXON_W, tr.getStatistic().intValue());
				out.put(WILCOXON_P, tr.getP());
				int diff = 0;
				for (int i=0; i<valsA.length; i++) if (valsA[i] <= valsB[i]) {diff++; } else {diff--;}
				out.put(WILCOXON_WINNER, targets.get(diff > 0 ? 0 : 1));
				if (stats.isPermutationTestSupported()) {
					//=== Permutation test : only implemented in RStatistics.
					tr = stats.permutationTest(valsA, valsB, Alternative.TWO_SIDED, true);
					out.put(PERMUTATION_P, tr.getP());
				} else {
					log.warning("Permutation test not supported");
				}
			}
			return out;
		}
		
	}
}
