package ca.ubc.cs.beta.smac;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.json.JSONObject;

import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.TransformedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.metaalgorithms.design.ConfiguratorImplementation;
import ca.ubc.cs.beta.hal.algorithms.parameters.Domain;
import ca.ubc.cs.beta.hal.algorithms.parameters.IntegerDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSetting;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSettingBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpace;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpaceBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.RealDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.SampleableDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.algorithms.transformations.LinearizingTransformation;
import ca.ubc.cs.beta.hal.analysis.Statistics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.ReadOnlyDataManager;
import ca.ubc.cs.beta.hal.environments.SubrunRunner;
import ca.ubc.cs.beta.hal.problems.InstanceList;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.MetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;


/***
 * Re-used CN's boiler plate, but re-implemented ROAR from scratch.
 * @author Frank Hutter
 *
 */
public class ROARConfiguratorImplementation extends ConfiguratorImplementation implements ImmutableJsonSerializable {

	public static final ParameterSpace DFLT_CFG_SPACE, DFLT_SCEN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE;
	public static final String NAME = "ROAR", VERSION="1.1.0";
	public static final Map<String, Object> PROPERTIES;
	private static final Logger log = Logger.getLogger(ROARConfiguratorImplementation.class.getCanonicalName());
	static {		
		ParameterSpaceBuilder psb = new ParameterSpaceBuilder();
		DFLT_CFG_SPACE = psb.build();
		psb.put(Semantics.MAX_CPUTIME, new RealDomain(0., null, 24*60*60.));
		psb.put(Semantics.SEED, new IntegerDomain(Integer.MIN_VALUE, Integer.MAX_VALUE, 0));
		DFLT_SCEN_SPACE = psb.build();

		psb.clear();
		psb.put(Semantics.DESIGN, Semantics.getDomain(Semantics.DESIGN));
		psb.put(Semantics.SOLUTION_QUALITY, Semantics.getDomain(Semantics.SOLUTION_QUALITY));
		DFLT_OUT_SPACE = psb.build();

		psb.clear();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, new RealDomain(0., null, 60.));
		SUPP_OPT_SPACE = psb.build();

		HashMap<String, Object> map = new HashMap<String, Object>();
		map.put(Semantics.DETERMINISTIC, false);
		map.put(Semantics.EXPORTABLE, true);
		map.put(Semantics.CUTOFF_AGNOSTIC, true);
		PROPERTIES = Collections.unmodifiableMap(map);		
	}

	public ROARConfiguratorImplementation() {
		super(NAME, VERSION, PROPERTIES, DFLT_CFG_SPACE,
				DFLT_SCEN_SPACE, DFLT_OUT_SPACE, SUPP_OPT_SPACE);
//		log.setLevel(Level.INFO);
	}

	@SuppressWarnings("unchecked")
	@Override
	public Set<Set<String>> getRequiredTags() {
		return Misc.asSet(Misc.asSet(MetaProblemInstance.CONFIGURABLE_TARGET_TAG,
				MetaProblemInstance.TAG, InstanceMetricMetaProblemInstance.TAG));
	}

	@Override
	public MetaAlgorithmRun getRun(AlgorithmRunRequest request,
			SubrunRunner runner, ReadOnlyDataManager datamanager,
			Statistics stats) {
		return new ROARMetaAlgorithmRun(request, runner, datamanager);
	}

	@Override
	public JSONObject buildSpec() {
		JSONObject out = super.buildSpec();
		out.put("version", VERSION);
		return out;
	}

	public static ROARConfiguratorImplementation fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(ROARConfiguratorImplementation.class, spec);
		String ver = o.optString("version");
		if (!VERSION.equals(ver) && ver != null && ver.length() > 0) {
			throw new IllegalArgumentException("Record corresponds to different version of ROAR plugin; was:" + 
					ver + "; is: " + VERSION );
		}
		return new ROARConfiguratorImplementation();
	}

	private static class ROARMetaAlgorithmRun extends MetaAlgorithmRun {

		private final Double maxcputime;
		private final ParameterizedAlgorithm targetcl;
		private final SubrunRunner runner;
		private final InstanceList instances;
		private volatile double targetAlgoTimeSpent = 0;
		private volatile int iteration = 0;
		private final PerformanceMetric metric;
		private final Random rand;
		private RunHistory runHistory;
		private ParameterSetting incumbent;
		private double currentIncumbentCost = 100000000; // not sure if HAL can handle infinity due to serialization
		Map<ProblemInstance, Random> instanceRNGs;
		private final SampleableDomain<? extends Number> seedDomain;

		/* Chris's boiler plate for setting up everything. */
		@SuppressWarnings("unchecked")
		public ROARMetaAlgorithmRun(AlgorithmRunRequest request, SubrunRunner runner, ReadOnlyDataManager dm) {
			super(request, runner, dm);
			runHistory = new RunHistory();
			
			InstanceMetricMetaProblemInstance inst = (InstanceMetricMetaProblemInstance) request.getProblemInstance();
			long seed = ((Number)request.getScenarioValue(Semantics.SEED)).longValue();
			rand = Global.getRandom(seed);

			instanceRNGs = new HashMap<ProblemInstance, Random>();
			for (ProblemInstance pi: inst.getInstanceDistribution()) {
				instanceRNGs.put(pi, Global.getRandom(rand.nextLong() + pi.hashCode()));
			}
			
			if (!(inst.getInstanceDistribution() instanceof InstanceList))
				throw new IllegalArgumentException("ROAR only supports Instance Lists");
			if (!(inst.getAlgorithms() instanceof ParameterizedAlgorithm))
				throw new IllegalArgumentException("ROAR only supports a single parameterized target algorithm");
			ParameterizedAlgorithm alg = ((ParameterizedAlgorithm) inst.getAlgorithms()).clone();

			metric = inst.getMetric();
			instances = (InstanceList) inst.getInstanceDistribution();

			if (!LinearizingTransformation.isLinear(alg)) {
				log.info("Linearizing algorithm domains");
				alg = new LinearizingTransformation().transform(alg);
			}
			Number captime = (Number) inst.getOption(Semantics.MAX_SUBRUN_CPUTIME);
			captime = (Number) (captime == null ? SUPP_OPT_SPACE.get(Semantics.MAX_SUBRUN_CPUTIME).getDefaultValue() : captime);
			alg.setScenarioValue( Semantics.MAX_CPUTIME, captime.doubleValue() );

			targetcl = alg.clone();
			incumbent = targetcl.getConfigurationSpace().getDefaults();
			seedDomain = (SampleableDomain<? extends Number>) (targetcl.hasScenarioVariable(Semantics.SEED) ? 
					targetcl.getScenarioDomain(Semantics.SEED): Semantics.getDomain(Semantics.SEED)); 

			maxcputime = ((Number)request.getScenarioValue(Semantics.MAX_CPUTIME)).doubleValue();
			this.runner = runner;
		}

		@Override
		protected void start() {
			setStatus(RunStatus.RUNNING);
			iteration = 1;
			try {
				//=== Initialize with one run using the default.
				ProblemInstance pi = sampleProblemInstanceUnifRand();
				long seed = seedDomain.getSample(instanceRNGs.get(pi)).longValue();
				evaluateRun(incumbent, pi, seed);
				
				//=== Iteratively select a new random configuration, and compare it to the incumbent.
//				while ( targetAlgoTimeSpent < maxcputime && getTerminationStatus() == null) {
				while ( getTotalCpuTime() < maxcputime && getTerminationStatus() == null) {
					//=== Select a challenger uniformly at random.
					//log.info("Iteration "+iteration+": Selecting challengers...");
					ArrayList<ParameterSetting> challengers = new ArrayList<ParameterSetting>();
					challengers.add(getRandomConfig());
	
					//=== Intensification.
					//log.info("Iteration "+iteration+": Intensifying...");
					intensify(challengers, 0);

					//=== Report the current incumbent, etc, to HAL.
					super.updateOutput(getOutputSnapshot());
					log.info("Iteration "+iteration+": " + targetAlgoTimeSpent + "/" + maxcputime + "s used. Incumbent " +incumbent+" with empirical cost "+currentIncumbentCost + ".");
					iteration++;
				}
				if (getTotalCpuTime() >= maxcputime ){
//				if( targetAlgoTimeSpent >= maxcputime ){
					log.info("Time limit is up.");
				} else {
					log.log(Level.WARNING, "External termination, termination status " + getTerminationStatus(), getException());
				}
				log.info("Final incumbent: "+incumbent+", with empirical cost "+currentIncumbentCost);
			} catch (Exception e0) {
				//=== On an unexpected exception, terminate.
				log.log(Level.SEVERE, "ROAR exception ", e0);
				terminate(e0);
			} finally {
				setStatus(getTerminationStatus() == null ? RunStatus.FINISHED : getTerminationStatus());
			}
		}

		@Override
		public double getFractionCompleted() {
			return Math.min(getTotalCpuTime() / maxcputime, 0.99);
		}
		
		@Override
		public synchronized Map<String, Object> getOutputSnapshot() {
			Map<String, Object> out = new HashMap<String, Object>();
			synchronized(targetcl) {
				targetcl.updateConfiguration(incumbent);
				// untransform the reported incumbent
				ParameterizedAlgorithm targ = targetcl;
				if (targ instanceof TransformedAlgorithm) targ = ((TransformedAlgorithm)targ).getCore();
				out.put(Semantics.DESIGN, targ.getParameterlessAlgorithm());
			}
			out.put(Semantics.SOLUTION_QUALITY, currentIncumbentCost);
			return out;
		}

		
		/* Get a new parameter configuration, sampling uniformly at random. */
		protected ParameterSetting getRandomConfig() {	
			Set<String> parameterNames = targetcl.getConfigurationSpace().keySet();
			ParameterSettingBuilder randomConfig = new ParameterSettingBuilder();
			
			// get a new value for each key
			for ( String key : parameterNames ) {
				Domain d = targetcl.getConfigurationDomain(key);
				if ( d instanceof SampleableDomain<?> ) {
					randomConfig.put( key, ((SampleableDomain<?>) d).getSample( rand ));      
				} else { 
					errorExit("Haven't implemented uniform random sampling for domain of type " + d.getClass().getCanonicalName());
				}
			}
			return targetcl.getConfigurationSpace().reduceSetting(randomConfig, true).build();
		}
		
		/* Execute the target algorithm with the specified config on the specified instance/seed combo,
		 * and return its performance. Also update the time we spent running target algos.  
		 * In the future, want to add captime (via "ParameterSetting scenario") */
		private double evaluateRun(ParameterSetting config, ProblemInstance pi, long seed) {
			targetcl.updateConfiguration(config);
			targetcl.setProblemInstance(pi);
			if( !targetcl.isDeterministic() ){
				targetcl.setScenarioValue( Semantics.SEED, seed );
			}
			AlgorithmRunRequest request = targetcl.getAlgorithmRunRequest();
			request.setDistributedExecutionOK(false);
			AlgorithmRun run = runner.fetchRun(request);
			Global.getThreadPool().execute(run);
			while (!RunStatus.isFinished(run.getStatus())) {
				try {
					run.waitFor();
				} catch (InterruptedException e) {
					// FIXME: dont ignore
				}
			}
			
			//=== Update the history, and the time spent.
			// FIXME: the metric might be null; yields an NPE
			double runCost = metric.evaluate(run);
			log.info("Logging config " + config + ", pi " + pi + ", seed " + seed + ", runCost " + runCost + " to runHistory.");
			runHistory.append(config, pi, seed, runCost);
			targetAlgoTimeSpent += run.getTotalCpuTime(); // FIXME: this incldues overheads; is it intentional?
			if (config.equals(incumbent)){
				updateIncumbentCost();
			}
			
			return runCost;
		}
		
		@SuppressWarnings("unused")
		private Map<InstanceSeedCombo, Double> evaluateRuns(ParameterSetting config, List<InstanceSeedCombo> torun) {
			targetcl.updateConfiguration(config);
			Map<InstanceSeedCombo, AlgorithmRun> runs = new HashMap<InstanceSeedCombo, AlgorithmRun>(torun.size());
			Map<InstanceSeedCombo, Double> out = new HashMap<InstanceSeedCombo, Double>(torun.size());
			// launch all the runs
			for (InstanceSeedCombo instseed : torun) {
				targetcl.setProblemInstance(instseed.getPi());
				if( !targetcl.isDeterministic() ){
					targetcl.setScenarioValue( Semantics.SEED, instseed.getSeed() );
				}
				AlgorithmRun run = runner.fetchRun(targetcl.getAlgorithmRunRequest());
				Global.getThreadPool().execute(run);
				runs.put(instseed, run);
			}
			// wait for all the runs to finish
			for (Entry<InstanceSeedCombo, AlgorithmRun> kv: runs.entrySet()) {
				AlgorithmRun run = kv.getValue();
				while (!RunStatus.isFinished(run.getStatus())) {
					try {
						run.waitFor();
					} catch (InterruptedException e) {
						// FIXME: dont ignore
					}
				}
				InstanceSeedCombo instseed = kv.getKey();
				ProblemInstance pi = instseed.getPi();
				long seed = instseed.getSeed();
				
				//=== Update the history, and the time spent.
				double runCost = metric.evaluate(run);
				log.info("Logging config " + config + ", pi " + pi + ", seed " + seed + ", runCost " + runCost + " to runHistory.");
				runHistory.append(config, pi, seed, runCost);
				targetAlgoTimeSpent += run.getTotalCpuTime(); // FIXME: this incldues overheads; is it intentional?
				
				if (config.equals(incumbent)){
					updateIncumbentCost();
				}
				out.put(instseed, runCost);
			}
			return out;
		}
		
		private static void errorExit(String errorMessage){
			System.err.println(errorMessage);
			System.exit(-1);
		}
		
		private ProblemInstance sampleProblemInstanceUnifRand(){
			return instances.get(rand.nextInt(instances.size()));
		}
		
		/* ROAR's and SMAC's intensify procedure. Performs runs for incumbent and challengers, and potentially changes the incumbent. */
		private void intensify(ArrayList<ParameterSetting> challengers, double timeBound){
			double initialTime = targetAlgoTimeSpent;
			for ( int i=0; i<challengers.size(); i++ ){
				//log.info("Challenger #" + (i+1) + ". Time spent in this call to intensify: " + (targetAlgoTimeSpent-initialTime) + " (stop intensify after : " + timeBound + ")");
				if ( targetAlgoTimeSpent-initialTime > timeBound && i >= 1 ) {
					break;
				}
				challengeIncumbent( challengers.get(i) );
			}
		}

		/* ROAR's and SMAC's challengeIncumbent procedure. Performs runs for incumbent and challenger, and potentially changes the incumbent. */
		private void challengeIncumbent(ParameterSetting challenger){
			//TODO: make maximum #runs a parameter of ROAR.
			//=== Perform run for incumbent unless it has the maximum #runs.
			if (runHistory.getTotalNumRunsOfConfig(incumbent) < 2000){
				log.info("One extra run for incumbent.");
				ProblemInstance pi = runHistory.getRandomInstanceWithFewestRunsFor(incumbent, instances, rand);
				long seed = seedDomain.getSample(instanceRNGs.get(pi)).longValue();
				evaluateRun(incumbent, pi, seed);
			}
			
			int N=1;
			while(true){
				//=== Get (instance,seed) combos missing for challenger.
				//Set<InstanceSeedCombo> sMissing = runHistory.getInstanceSeedCombosRan(incumbent);
				Set<InstanceSeedCombo> sMissing = new HashSet<InstanceSeedCombo>( runHistory.getInstanceSeedCombosRan(incumbent) );
				//System.out.println("sInc: " + sMissing);
				sMissing.removeAll( runHistory.getInstanceSeedCombosRan(challenger) );
				//System.out.println("sChal: " +  runHistory.getInstanceSeedCombosRan(challenger));
				
				//System.out.println("sMissing is the difference: " + sMissing);
				//System.out.println("Size of sMissing: " + sMissing.size());
				
				//=== Shuffle them, and perform a run of the challenger for the first N.
				ArrayList<InstanceSeedCombo> aMissing = new ArrayList<InstanceSeedCombo>();
				aMissing.addAll(sMissing);
				Collections.shuffle(aMissing);
				log.info("Performing up to " + N + " runs for challenger.");
				
				// CN: below 3 lines do this step in parallel
//				List<InstanceSeedCombo> torun = aMissing.subList(0, Math.min(N, aMissing.size()));
//				evaluateRuns(challenger, torun);
//				sMissing.removeAll(torun);
				
				// CN: below 6 lines do this step sequentially
				for (int i = 0; i < Math.min(N, aMissing.size()); i++) {
					evaluateRun(challenger, aMissing.get(i).getPi(), aMissing.get(i).getSeed());
					sMissing.remove(aMissing.get(i));
					log.info("sMissing after removal: " + sMissing);
					log.info("Size of sMissing after removal: " + sMissing.size());
				}
				
				//=== Get performance of incumbent and challenger on their common instances.
				Set<ProblemInstance> piCommon = runHistory.getInstancesRan(incumbent);
				piCommon.retainAll( runHistory.getInstancesRan( challenger ));
				
				double incCost = runHistory.getEmpiricalCost(incumbent, metric, piCommon);
				double chalCost = runHistory.getEmpiricalCost(challenger, metric, piCommon);
				
				//=== Decide whether to discard challenger, to make challenger incumbent, or to continue evaluating it more.
				if (incCost < chalCost){
					log.info("Challenger (" + chalCost + ") worse than incumbent (" + incCost + "). Discarding it.");
					break;
				}
				if (sMissing.isEmpty()){
					log.info("Challenger (" + chalCost + ") at least as good as incumbent based on all its runs. Making challenger new incumbent.");
					changeIncumbentTo(challenger);
					break;
				}
				N *= 2;
			}
		}

		//=== Update the incumbent's empirical cost.
		private void updateIncumbentCost(){
			currentIncumbentCost = runHistory.getEmpiricalCost(incumbent, metric, new HashSet<ProblemInstance>(instances));
		}
		
		//=== Set the incumbent to the input configuration, and update its empirical cost.
		private void changeIncumbentTo(ParameterSetting newIncumbent){
			incumbent = newIncumbent;
			updateIncumbentCost();
			log.info("New incumbent: " + incumbent + ".\nEmpirical cost based on " + runHistory.getInstanceSeedCombosRan(incumbent).size() + " runs: " + currentIncumbentCost + ".");
		}

	}
}
