package ca.ubc.cs.beta.smac;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSetting;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.problems.InstanceList;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;

public class RunHistory {
	private final ArrayList<ProblemInstance> pisUsed = new ArrayList<ProblemInstance>();
	private final ArrayList<ParameterSetting> configsUsed = new ArrayList<ParameterSetting>();
	private final Set<Long> seedsUsed = new HashSet<Long>();
	private final ArrayList<Double> runCosts = new ArrayList<Double>();

	private final HashMap<ParameterSetting, Integer> numRunsForConfig = new HashMap<ParameterSetting, Integer>();
	private final HashMap<ParameterSetting, HashMap<ProblemInstance, HashMap<Long, Double>>> configToPerformanceMap =
		new HashMap<ParameterSetting, HashMap<ProblemInstance, HashMap<Long, Double>>>();

	/* Add the runtime data for a single target algorithm run to the history. */
	public void append(ParameterSetting config, ProblemInstance pi, long seed, double runCost){
		//=== Update full map of config, instance, seed, and runCost.
		if (!configToPerformanceMap.containsKey(config)){
			configToPerformanceMap.put(config, new HashMap<ProblemInstance, HashMap<Long, Double>>());
		}
		HashMap<ProblemInstance, HashMap<Long, Double>> hConfig = configToPerformanceMap.get(config);

		if (!hConfig.containsKey(pi)){
			hConfig.put(pi, new HashMap<Long, Double>());
		}
		HashMap<Long, Double> hConfigInst = hConfig.get(pi);

		if (hConfigInst.containsKey(seed)){
			throw new RuntimeException("ERROR in control logic: we already logged a run with identical config, instance and seed.");
		} else {
			hConfigInst.put(seed, runCost);
		}
		
		//=== Increase counter for total #runs for config.
		if (!numRunsForConfig.containsKey(config)){
			numRunsForConfig.put(config, 0);
		}
		numRunsForConfig.put(config, numRunsForConfig.get(config).intValue()+1 );

		//=== Update single array lists.
		configsUsed.add(config);
		pisUsed.add(pi);
		seedsUsed.add(seed);
		runCosts.add(runCost);
	}

	/* Return instances we have already run for config. */
	public Set<ProblemInstance> getInstancesRan(ParameterSetting config){
		if (!configToPerformanceMap.containsKey(config)){
			return new HashSet<ProblemInstance>();
		}
		return new HashSet<ProblemInstance>( configToPerformanceMap.get(config).keySet() );
	}

	/* Return <instance,seed> combos we have already run for config. */
	public Set<InstanceSeedCombo> getInstanceSeedCombosRan(ParameterSetting config){
		if (!configToPerformanceMap.containsKey(config)){
			return new HashSet<InstanceSeedCombo>();
		}
		Set<InstanceSeedCombo> instanceSeedCombos = new HashSet<InstanceSeedCombo>();
		HashMap<ProblemInstance, HashMap<Long, Double>> hConfig = configToPerformanceMap.get(config);
		for (Entry<ProblemInstance, HashMap<Long, Double>> kv : hConfig.entrySet()) {
			//=== For each instance we have run for the config, add a pair for each of the run seeds.
			ProblemInstance pi =  kv.getKey();
			HashMap<Long, Double> hConfigInst = kv.getValue();
			for (Long seed: hConfigInst.keySet()) {
				instanceSeedCombos.add( new InstanceSeedCombo(pi, seed) );
			}
		}
		return instanceSeedCombos;
	}
	
	/* Compute and return the empirical cost for a parameter configuration on the subset of the provided instance set that we have runs for. */
	public double getEmpiricalCost(ParameterSetting config, PerformanceMetric metric, Set<ProblemInstance> instanceSet){
		if (!configToPerformanceMap.containsKey(config)){
			return Double.MAX_VALUE;
		}
		ArrayList<Double> instanceCosts = new ArrayList<Double>();
		HashMap<ProblemInstance, HashMap<Long, Double>> hConfig = configToPerformanceMap.get(config);

		//=== Instances to use: intersection of input instance set and instances ran.
		Set<ProblemInstance> instancesToUse = new HashSet<ProblemInstance>();
		instancesToUse.addAll(instanceSet);
		instancesToUse.retainAll(hConfig.keySet());

		for (Iterator<ProblemInstance> iterator = instancesToUse.iterator(); iterator.hasNext();) {
			ProblemInstance pi = (ProblemInstance) iterator.next();
			HashMap<Long, Double> hConfigInst = hConfig.get(pi);

			//=== Aggregate costs for each of the seeds we ran for the instance.
			ArrayList<Double> localCosts = new ArrayList<Double>();
			for (Iterator<Long> iterator2 = hConfigInst.keySet().iterator(); iterator2.hasNext();) {
				long seed = (Long) iterator2.next();
				localCosts.add( hConfigInst.get(seed) );
			}
			instanceCosts.add( metric.aggregate( localCosts )); 
		}

		//=== In the end, aggregate costs across the instances.
		return metric.aggregate(instanceCosts);
	}

	/* Get a random element of the set of instances for which config has the fewest runs. */
	public ProblemInstance getRandomInstanceWithFewestRunsFor(ParameterSetting config, InstanceList instanceList, Random rand){
		HashMap<ProblemInstance, HashMap<Long, Double>> hConfig = configToPerformanceMap.get(config);
		
		//=== Make every instance that hasn't been run yet for config a candidate.
		Set<ProblemInstance> candidates = new HashSet<ProblemInstance>();
		candidates.addAll(instanceList);
		if (configToPerformanceMap.containsKey(config)){
			candidates.removeAll(hConfig.keySet());
		}
		
		//=== If all instances have at least one run, go through and get the instances with minimal #runs.
		if (candidates.size() == 0){
			int minNumRuns = Integer.MAX_VALUE;
			for (Iterator<ProblemInstance> iterator = instanceList.iterator(); iterator.hasNext();) {
				ProblemInstance inst = (ProblemInstance) iterator.next();
				int numRuns = hConfig.get(inst).size();
				if (numRuns <= minNumRuns){
					if (numRuns < minNumRuns){ // new value for fewest runs -> ditch all previous candidates
						candidates.clear();
						minNumRuns = numRuns;
					}
					candidates.add(inst);
				}
			}
		}
			
		//=== Return a random element of the candidate instance set (it's sad there is no method for that in Java's Set).
		return (ProblemInstance) candidates.toArray()[rand.nextInt(candidates.size())];
	}
	
	public int getTotalNumRunsOfConfig(ParameterSetting config){
		if (numRunsForConfig.containsKey(config)){
			return numRunsForConfig.get(config);
		} else {
			return 0;
		}
	}
	
}
