package ca.ubc.cs.beta.smac.surrogates;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.commons.io.FileUtils;

import com.mathworks.toolbox.javabuilder.*;

import net.sf.json.JSONArray;
import net.sf.json.JSONObject;

import ca.ubc.cs.beta.activeconfigurator.matlab.predictivemodels.PredictiveModel;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.AlgorithmRunRequest;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.parameters.DiscreteDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.Domain;
import ca.ubc.cs.beta.hal.algorithms.parameters.NumericalDomain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSetting;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSpace;
import ca.ubc.cs.beta.hal.models.PerformanceModel;
import ca.ubc.cs.beta.hal.problems.Feature;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.utils.GZIPStringCompressor;
import ca.ubc.cs.beta.hal.utils.ImmutableJsonSerializable;
import ca.ubc.cs.beta.hal.utils.Misc;
import ca.ubc.cs.beta.hal.utils.StringCompressor;

public class MatlabAlgorithmPerformanceModel implements PerformanceModel, ImmutableJsonSerializable {

	private final ParameterizedAlgorithm target;
	private final List<String> parameterOrder;
	private final List<Feature> featureOrder;
	private final Map<String, List<Object>> domainValueOrders;
	private final JsonHelper<MatlabAlgorithmPerformanceModel> helper = new JsonHelper<MatlabAlgorithmPerformanceModel>(this);
	private final PredictiveModel matlabModel;
	private final Object model;
	private final byte[] matlabData;
	
	private static final StringCompressor COMPRESSOR = new GZIPStringCompressor();
	
	private static final Logger log = Logger.getLogger(MatlabAlgorithmPerformanceModel.class.getCanonicalName());
	
	private static File writeData(byte[] modelData) {
		try {
			File f = File.createTempFile("matlab", ".model");
			FileUtils.writeByteArrayToFile(f, modelData);
			return f;
		} catch (IOException e) {
			throw new RuntimeException("Error writing matlab data", e);
		}
	}
	
	private static File writeParameterOrder(List<String> parameterOrder) {
		try {
			File f = File.createTempFile("matlab", ".model");
			FileUtils.writeLines(f, parameterOrder);
			return f;
		} catch (IOException e) {
			throw new RuntimeException("Error writing parameter order", e);
		}
	}
	
	private static byte[] readData(File modelFile) {
		try {
			return FileUtils.readFileToByteArray(modelFile);
		} catch (IOException e1) {
			throw new RuntimeException("Can't read matlab model: " + modelFile, e1);
		}
	}
	
	@SuppressWarnings("unchecked")
	private static List<String> readParameterOrder(File parameterOrderFile) {
		try {
			return FileUtils.readLines(parameterOrderFile);
		} catch (IOException e1) {
			throw new RuntimeException("Can't read matlab parameter order: " + parameterOrderFile, e1);
		}
	}
	
	public MatlabAlgorithmPerformanceModel(ParameterizedAlgorithm target, 
			File modelFile,
			File parameterOrderFile,
			List<Feature> featureOrder, 
			Map<String, List<Object>> domainValueOrders) {
		this(target, modelFile, parameterOrderFile, 
				readData(modelFile), readParameterOrder(parameterOrderFile),
				featureOrder, domainValueOrders);
	}
	
	private MatlabAlgorithmPerformanceModel(ParameterizedAlgorithm target, 
			byte[] matlabData,
			List<String> parameterOrder,
			List<Feature> featureOrder, 
			Map<String, List<Object>> domainValueOrders) {
		this(target, writeData(matlabData), writeParameterOrder(parameterOrder), 
				matlabData, parameterOrder,
				featureOrder, domainValueOrders);
	}
	
	private MatlabAlgorithmPerformanceModel(ParameterizedAlgorithm target, 
			File modelFile,
			File parameterOrderFile,
			byte[] matlabData, 
			List<String> parameterOrder,
			List<Feature> featureOrder, 
			Map<String, List<Object>> domainValueOrders) {
		this.target = target;
		this.featureOrder = featureOrder;
		this.domainValueOrders = domainValueOrders;
		this.parameterOrder = parameterOrder;
		this.matlabData = matlabData;
		
		// initialize matlab connection and objects
		try {
			matlabModel = new PredictiveModel();
			model = matlabModel.read_model_from_file(1, modelFile.getPath());
		} catch (MWException e) {
			throw new RuntimeException("Error connecting to Matlab", e);
		}
		
		//validate features
		for (Feature f: featureOrder) {
			if (!Number.class.isAssignableFrom(f.getDomain().getContainedClass())) {
				throw new IllegalArgumentException("Only numerically-valued features accepted");
			}
		}
		
		//validate parameter domains
		for ( Entry<String, Domain> kv : target.getConfigurationSpace().entrySet()) {
			Domain d = kv.getValue();
			if (domainValueOrders.containsKey(kv.getKey()) || d instanceof NumericalDomain) continue;
			if (d instanceof DiscreteDomain) {
				log.warning("Assuming the domain ordering in " + kv.getKey() + " matches that used in Matlab.  " +
						"Consider explicitly specifying the domain order!");
				continue;
			}
			throw new IllegalArgumentException("Cannot infer numeric representation for domain " + kv.getKey());
		}
	}
	
	private double[] convertParameterSetting(ParameterSetting ps) {
		double[] out = new double[parameterOrder.size()];
		int index = 0;
		for (String s: parameterOrder) {
			Domain d = target.getConfigurationDomain(s);
			if (d instanceof NumericalDomain) {
				out[index++] = ((Number)ps.get(s)).doubleValue();
			} else if (domainValueOrders.containsKey(s)) {
				out[index++] = Integer.valueOf(domainValueOrders.get(s).indexOf(ps.get(s)));
			} else if (d instanceof DiscreteDomain) {
				int i=0;
				Object o = ps.get(s);
				for (Object o2: (DiscreteDomain<?>)d) {
					if (o.equals(o2)) break;
					i++;
				}
				out[index++] = i;
			}
			else {
				throw new IllegalArgumentException("Can't convert value for parameter " + s + " into a number for Matlab");
			}
		}
		return out;
	}
	
	private double[] convertInstanceFeatures(ProblemInstance inst) {
		double[] out = new double[featureOrder.size()];
		int index = 0;
		for (Feature f: featureOrder) {
			Number n = (Number) inst.getFeatureValue(f);
			if (n == null) throw new RuntimeException("Instance " + inst + " missing requred feature " + f);
			out[index++] = n.doubleValue();
		}
		return out;
	}
	
	@Override
	protected void finalize() throws Throwable {
		if (matlabModel != null) matlabModel.dispose();
		super.finalize();
	}
	
	@Override
	public String toSpec() {
		return helper.getSpec();
	}

	@Override
	public String getHash() {
		return helper.getHash();
	}

	@Override
	public boolean equals(Object o) {
		return helper.getEquals(o);
	}
	
	@Override
	public int hashCode() {
		return helper.getHashCode();
	}
	
	@Override
	public ResponseStatistics predictStatistics(AlgorithmRunRequest... requests) {
		
		double[][] configs = new double[requests.length][parameterOrder.size()], 
					insts = new double[requests.length][featureOrder.size()];
		int i=0;
		for (AlgorithmRunRequest req: requests) {
			if (!req.getImplementation().equals(target.getImplementation())) {
				throw new IllegalArgumentException("Request " + req + " doesnt match expected algorithm " + target);
			}
			configs[i] = convertParameterSetting(req.getConfiguration());
			insts[i] = convertInstanceFeatures(req.getProblemInstance());
		}

		// TODO: matlab call
		
		return null;
	}

	@Override
	public EmpiricalDistribution predictDistribution(AlgorithmRunRequest... requests) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public JSONObject buildSpec() {
		JSONObject spec = helper.getStubSpec();
		spec.put("target", target.toSpec());
		spec.put("parameterOrder", this.parameterOrder);
		JSONArray arr = new JSONArray();
		for (Feature f: featureOrder) arr.add(f.toSpec());
		spec.put("featureOrder", arr);
		JSONObject jo = new JSONObject();
		List<String> keys = new ArrayList<String>(domainValueOrders.keySet());
		Collections.sort(keys);
		for (String key: keys) {
			JSONArray ja = new JSONArray();
			for (Object o: domainValueOrders.get(key)) {
				ja.add(JsonHelper.serializeValue(o));
			}
			jo.put(key, ja);
		}
		spec.put("valueOrders", jo);
		spec.put("matlabData", COMPRESSOR.compressToString(matlabData));
		return spec;
	}
	
	public static MatlabAlgorithmPerformanceModel fromSpec(String spec) {
		JSONObject o = JsonHelper.readSpecStub(MatlabAlgorithmPerformanceModel.class, spec);
		ParameterizedAlgorithm target = (ParameterizedAlgorithm) Misc.fromSpec(o.getString("target"));
		List<String> parameterOrder = JsonHelper.jsonArrayToList(o.getJSONArray("parameterOrder"));
		List<Feature> featureOrder = new LinkedList<Feature>();
		JSONArray arr = o.getJSONArray("featureOrder");
		for (int i=0; i<arr.size(); i++) {
			featureOrder.add((Feature) Misc.fromSpec(arr.getString(i)));
		}
		Map<String, List<Object>> valueOrders = new HashMap<String, List<Object>>();
		JSONObject vo = o.getJSONObject("valueOrders");
		for (Object ko: vo.keySet()) {
			JSONArray ja = vo.getJSONArray((String)ko);
			List<Object> objs = new ArrayList<Object>(ja.size());
			for (int i=0; i<ja.size(); i++) {
				objs.add(ja.get(i));
			}
			valueOrders.put((String)ko, objs); 
		}
		byte[] matlabData = COMPRESSOR.decompressByte(o.getString("matlabData"));
		return new MatlabAlgorithmPerformanceModel(target, matlabData, parameterOrder, featureOrder, valueOrders);
	}

	@Override
	public Set<Feature> getRequiredInstanceFeatures() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public ParameterSpace getRequiredAlgorithmConfigurationSpace() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public ParameterSpace getRequiredAlgorithmScenarioSpace() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Set<AlgorithmImplementation> getSupportedAlgorithmImplementations() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public String toFullSpec() {
		return toSpec();
	}

}
