package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import static ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis.AnalysisSemantics.*;
import static org.junit.Assert.*;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;

import org.apache.commons.io.FileUtils;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import ca.ubc.cs.beta.hal.algorithms.Algorithm;
import ca.ubc.cs.beta.hal.algorithms.MetaAlgorithmImplementation;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithmList;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSettingBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun.RunStatus;
import ca.ubc.cs.beta.hal.environments.Environment;
import ca.ubc.cs.beta.hal.environments.Environment.Preference;
import ca.ubc.cs.beta.hal.environments.ExecutionManager.OutputHandlingOption;
import ca.ubc.cs.beta.hal.environments.FullAccessDataManager;
import ca.ubc.cs.beta.hal.environments.datamanagers.InvalidURIException;
import ca.ubc.cs.beta.hal.environments.datamanagers.SQLiteDataManager;
import ca.ubc.cs.beta.hal.environments.executionmanagers.LocalExecutionManager;
import ca.ubc.cs.beta.hal.problems.FileProblemInstance;
import ca.ubc.cs.beta.hal.problems.InstanceList;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.MetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.Tag;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;


public class ComprehensivePairwiseComparisonTest {
	
	private static FullAccessDataManager dm, dm0;
	
	@BeforeClass
	public static void setup() throws InvalidURIException {
		Global.getBaseLogger().setLevel(Level.WARNING);
		dm0 = Global.getDataManager();
		dm = new SQLiteDataManager(URI.create("jdbc:sqlite:test/paircmp.db"));
		Global.setDataManager(dm);
	}

	@Before
	public void clearDB() {
		dm.clear();
	}
	
	@AfterClass
	public static void teardown() {
		Global.setDataManager(dm0);
		new File("test/paircmp.db").delete();
	}
	
	static Algorithm getTestAlgorithm() throws IOException {
		// get the meta-algorithm
		MetaAlgorithmImplementation alg = new ComprehensivePairwiseComparison();
		Algorithm analyser = alg.getDefaultParameterizedAlgorithm();
		
		String json = "{'command':'python', 'name':'algA', 'tags':[['dummy']], 'outputFormat':{'stdout':['foroutput $v$ $t$']}, 'outputs':{'v':{'domain':'Real()'}, 't':{'domain':'Real()'}}, 'path':'test', 'deterministic':false, 'inputFormat':{'callstring':['test.py $time$ [$hitime$]']}, 'inputs':{'time':{'domain':'Real[0, inf)', 'default':0.5}, 'hitime':{'domain':'Real[0, inf)', 'default':0, 'conditional':[{'hitime':{'domain':'Real(0, inf)'}}]}}}";
		ParameterizedAlgorithm alg1 = ParameterizedAlgorithm.fromSpec(json);
		
		LinkedList<ParameterlessAlgorithm> algl = new LinkedList<ParameterlessAlgorithm>();
		alg1.setConfigurationValue("time", 0.1);
		alg1.setConfigurationValue("hitime", 0.2);
		ParameterlessAlgorithm expectedWinner = alg1.getParameterlessAlgorithm();
		algl.add(expectedWinner);
		alg1.setConfigurationValue("time", 0.05);
		alg1.setConfigurationValue("hitime", 0.5);
		algl.add(alg1.getParameterlessAlgorithm());
		ParameterlessAlgorithmList algs = new ParameterlessAlgorithmList(algl);
		System.out.println(">" + algl.get(0).getName() + " " +  algl.get(1).getName());
		System.out.println(">" + algl.get(0).getConfiguration() + " " +  algl.get(1).getName());
		
		PerformanceMetric metric = PerformanceMetric.PAR10;
		
		// create some dummy instances for our dummy algorithm
		List<ProblemInstance> instances = new LinkedList<ProblemInstance>();
		for (int i=0; i<10; i++) {
			File f = File.createTempFile("dummy", ".inst");
			FileUtils.writeStringToFile(f, f.getName());
			f.deleteOnExit();
			instances.add(new FileProblemInstance(f));
			instances.add(new FileProblemInstance(f));
		}
		InstanceList il = new InstanceList("dummylist",instances);
		il.addTagsToAllInstances(Tag.DUMMY);
				
		ParameterSettingBuilder psb = new ParameterSettingBuilder();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, 0.3);
		psb.put(Semantics.RUNS_PER_INSTANCE, 2);
		MetaProblemInstance inst = new InstanceMetricMetaProblemInstance(algs, il, metric, "dummyAnalysisProb", psb.build());
		
		analyser.setProblemInstance(inst);
//		analyser.setScenarioValue(Semantics.MAX_CPUTIME, 10);
//		analyser.setScenarioValue(Semantics.MAX_RUNLENGTH, 40);
		return analyser;
	}
	
	@Test
	public void testRunsLocal() throws Exception {

		final Environment env = new Environment("local unittest", new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.DISCARD), Preference.MAX_SIMULT_TARGET_RUNS, 4);//, Environment.Preference.GLOBAL_DB_CPU_THRESH, 0.0);
				
		Algorithm analyser = getTestAlgorithm();
		Algorithm expectedWinner = ((MetaProblemInstance)analyser.getProblemInstance()).getAlgorithms().get(0);
			
		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());
		Global.getThreadPool().execute(run);
		
		while (!RunStatus.isFinished(run.getStatus())) {
			System.out.println(run.getFractionCompleted());
			Thread.sleep(500);
		}
		System.out.println(run.getFractionCompleted());

		run.waitForVisitors();
		assertTrue(RunStatus.finishedWithoutError(run.getStatus()));
		assertEquals(run.getSubrunMeasuredCpuTime(), dm.getRun(run.getId()).getSubrunMeasuredCpuTime(), 0.01);
		
//		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
//		p.display((Plot)run.getLastOutputValueOnly(SCD_1));
//		p.display((Plot)run.getLastOutputValueOnly(SCD_2));
//		p.display((Plot)run.getLastOutputValueOnly(SCD_OVERLAY));
//		p.display((Plot)run.getLastOutputValueOnly(SCATTER));
		
		assertEquals(expectedWinner, run.getLastOutputValueOnly(WILCOXON_WINNER));
		assertFalse(run.getLastOutputValueOnly(Q10_1).equals(run.getLastOutputValueOnly(Q90_1)));
		assertFalse(run.getLastOutputValueOnly(Q10_1).equals(run.getLastOutputValueOnly(Q10_2)));
		
		// confident about wilcoxon winner at 80% level
		assertEquals(0., (Double) run.getLastOutputValueOnly(WILCOXON_P), 0.2);
		// don't think there's significant correlation
		assertEquals(0., (Double) run.getLastOutputValueOnly(SPEARMAN_RHO), 0.5);
		// not confident about lack of correlation at better than 95% level
		assertEquals(1., (Double) run.getLastOutputValueOnly(SPEARMAN_P), 0.95);
	}
	
	
	@Test
	public void testRunsLocalDuplicateAlgorithm() throws Exception {

		final Environment env = new Environment("local unittest", new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.DISCARD), Preference.MAX_SIMULT_TARGET_RUNS, 4);//, Environment.Preference.GLOBAL_DB_CPU_THRESH, 0.0);
				
		Algorithm analyser = getTestAlgorithm();
		InstanceMetricMetaProblemInstance inst = (InstanceMetricMetaProblemInstance)analyser.getProblemInstance();
		ParameterlessAlgorithm expectedWinner = inst.getAlgorithms().get(0);
		inst = new InstanceMetricMetaProblemInstance(
				new ParameterlessAlgorithmList(expectedWinner, expectedWinner),
				inst.getInstanceDistribution(),
				inst.getMetric(),
				"duplicate target alg inst",
				inst.getOptions());
		analyser.setProblemInstance(inst);
			
		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());

		Global.getThreadPool().execute(run);
		
		while (!RunStatus.isFinished(run.getStatus())) {
			System.out.println(run.getFractionCompleted());
			Thread.sleep(500);
		}
		System.out.println(run.getFractionCompleted());

		run.waitForVisitors();
		assertTrue(RunStatus.finishedWithoutError(run.getStatus()));
		assertEquals(run.getSubrunMeasuredCpuTime(), dm.getRun(run.getId()).getSubrunMeasuredCpuTime(), 0.01);
		
//		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
//		p.display((Plot)run.getLastOutputValueOnly(SCD_1));
//		p.display((Plot)run.getLastOutputValueOnly(SCD_2));
//		p.display((Plot)run.getLastOutputValueOnly(SCD_OVERLAY));
//		p.display((Plot)run.getLastOutputValueOnly(SCATTER));
		
		assertEquals(expectedWinner, run.getLastOutputValueOnly(WILCOXON_WINNER));
		assertFalse(run.getLastOutputValueOnly(Q10_1).equals(run.getLastOutputValueOnly(Q90_1)));
		assertTrue(run.getLastOutputValueOnly(Q10_1).equals(run.getLastOutputValueOnly(Q10_2)));
		
		// confident about wilcoxon winner at 80% level
		assertEquals(0., (Double) run.getLastOutputValueOnly(WILCOXON_P), 0.2);
		// think there's significant correlation
		assertEquals(1.0, (Double) run.getLastOutputValueOnly(SPEARMAN_RHO), 0.5);
		// confident about lack of correlation at better than 95% level
		assertEquals(0.0, (Double) run.getLastOutputValueOnly(SPEARMAN_P), 0.05);

	}
	
	@Test
	public void testRunsLocalTerminate() throws Exception {
		
		LocalExecutionManager xm = new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.ECHO);
		
		final Environment env = new Environment("local unittest", xm, Environment.Preference.MAX_SIMULT_TARGET_RUNS, 10);
		Algorithm analyser = getTestAlgorithm();

		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());
		Global.getThreadPool().execute(run);
		double time = 5.8;
		Thread.sleep((long)(1000*time));
		run.terminate();
		run.waitForVisitors();
		double stat = run.getStatus();
		assertTrue(RunStatus.isFinished(stat));
		assertTrue(RunStatus.isTerminated(stat));
		assertTrue(run.getSubrunCount() <= ((Number)analyser.getScenarioValue(Semantics.MAX_RUNLENGTH)).longValue());
		assertTrue(run.getSubrunCount() > 0);
		assertEquals(time, run.getMeasuredCpuTime(), time*.9); // just needs to be in the right ballpark, and nonzero
	}
}
