package ca.ubc.cs.beta.hal.algorithms.metaalgorithms.analysis;

import static org.junit.Assert.*;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.logging.Level;

import org.apache.commons.io.FileUtils;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import ca.ubc.cs.beta.hal.algorithms.Algorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.ParameterlessAlgorithmList;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSettingBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.Semantics;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun.RunStatus;
import ca.ubc.cs.beta.hal.environments.Environment;
import ca.ubc.cs.beta.hal.environments.Environment.Preference;
import ca.ubc.cs.beta.hal.environments.ExecutionManager.OutputHandlingOption;
import ca.ubc.cs.beta.hal.environments.FullAccessDataManager;
import ca.ubc.cs.beta.hal.environments.datamanagers.InvalidURIException;
import ca.ubc.cs.beta.hal.environments.datamanagers.SQLiteDataManager;
import ca.ubc.cs.beta.hal.environments.executionmanagers.LocalExecutionManager;
import ca.ubc.cs.beta.hal.problems.FileProblemInstance;
import ca.ubc.cs.beta.hal.problems.InstanceList;
import ca.ubc.cs.beta.hal.problems.InstanceMetricMetaProblemInstance;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.Tag;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;

public class KWayComparisonTest {
	
	private static FullAccessDataManager dm, dm0;
	
	@BeforeClass
	public static void setup() throws InvalidURIException {
		Global.getBaseLogger().setLevel(Level.WARNING);
		dm0 = Global.getDataManager();
		dm = new SQLiteDataManager(URI.create("jdbc:sqlite:test/kcmp.db"));
		Global.setDataManager(dm);
	}

	@Before
	public void clearDB() {
		dm.clear();
	}
	
	@AfterClass
	public static void teardown() {
		Global.setDataManager(dm0);
		new File("test/kcmp.db").delete();
	}
	
	static InstanceMetricMetaProblemInstance getTestInstance() throws IOException {
		// get the meta-algorithm
		
		String json = "{'command':'python', 'name':'algA', 'tags':[['dummy']], 'outputFormat':{'stdout':['foroutput $v$ $t$']}, 'outputs':{'v':{'domain':'Real()'}, 't':{'domain':'Real()'}}, 'path':'test', 'deterministic':false, 'inputFormat':{'callstring':['test.py $time$ [$hitime$]']}, 'inputs':{'time':{'domain':'Real[0, inf)', 'default':0.5}, 'hitime':{'domain':'Real[0, inf)', 'default':0, 'conditional':[{'hitime':{'domain':'Real(0, inf)'}}]}}}";
		ParameterizedAlgorithm alg1 = ParameterizedAlgorithm.fromSpec(json);
		
		LinkedList<ParameterlessAlgorithm> algl = new LinkedList<ParameterlessAlgorithm>();
		alg1.setConfigurationValue("time", 0.1);
		alg1.setConfigurationValue("hitime", 0.2);
		ParameterlessAlgorithm expectedWinner = alg1.getParameterlessAlgorithm();
		algl.add(expectedWinner);
		alg1.setConfigurationValue("time", 0.12); // should be similar
		alg1.setConfigurationValue("hitime", 0.18);
		algl.add(alg1.getParameterlessAlgorithm());
		alg1.setConfigurationValue("time", 0.05); // should be similar
		alg1.setConfigurationValue("hitime", 0.25);
		algl.add(alg1.getParameterlessAlgorithm());
		alg1.setConfigurationValue("time", 0.15); // should be slightly worse
		alg1.setConfigurationValue("hitime", 0.20);
		algl.add(alg1.getParameterlessAlgorithm());
		alg1.setConfigurationValue("time", 0.05); // should be clearly worse
		alg1.setConfigurationValue("hitime", 0.5);
		algl.add(alg1.getParameterlessAlgorithm());
		ParameterlessAlgorithmList algs = new ParameterlessAlgorithmList(algl);
		System.out.println(">" + algl.get(0).getName() + " " +  algl.get(1).getName());
		System.out.println(">" + algl.get(0).getConfiguration() + " " +  algl.get(1).getName());
		
		PerformanceMetric metric = PerformanceMetric.PAR10;
		
		// create some dummy instances for our dummy algorithm
		LinkedList<ProblemInstance> il = new LinkedList<ProblemInstance>();
		for (int i=0; i<10; i++) {
			File f = File.createTempFile("dummy", ".inst");
			FileUtils.writeStringToFile(f, f.getName());
			f.deleteOnExit();
			FileProblemInstance fpi = new FileProblemInstance(f);
			il.add(fpi);
			il.add(fpi); // intentional duplication
		}
		InstanceList instances = new InstanceList("dummylist", il);
		instances.addTagsToAllInstances(Tag.DUMMY);

		ParameterSettingBuilder psb = new ParameterSettingBuilder();
		psb.put(Semantics.MAX_SUBRUN_CPUTIME, 0.3);
		psb.put(Semantics.RUNS_PER_INSTANCE, 2);
		InstanceMetricMetaProblemInstance inst = new InstanceMetricMetaProblemInstance(algs, instances, metric, "dummyAnalysisProb", psb.build());
		
		return inst;
	}
	
	@Test
	public void testRunsLocal() throws Exception {

		final Environment env = new Environment("local unittest", new LocalExecutionManager(OutputHandlingOption.DISCARD, OutputHandlingOption.DISCARD),
				Preference.MAX_SIMULT_TARGET_RUNS, 4);//, Environment.Preference.GLOBAL_DB_CPU_THRESH, 0.0);
				
		Algorithm analyser = new KWayComparison().getDefaultParameterizedAlgorithm();
		analyser.setProblemInstance(getTestInstance());
		analyser.setScenarioValue(AnalysisSemantics.MIN_OUTPUT_INTERVAL, 1.0);
		
		AlgorithmRun run = env.fetchRun(analyser.getAlgorithmRunRequest());
		Global.getThreadPool().execute(run);
		
		while (!RunStatus.isFinished(run.getStatus())) {
			System.out.println(run.getFractionCompleted());
			Thread.sleep(500);
		}
		System.out.println(run.getFractionCompleted());

		run.waitForVisitors();
		assertTrue(RunStatus.isFinished(run.getStatus()));
		assertEquals(run.getSubrunMeasuredCpuTime(), dm.getRun(run.getId()).getSubrunMeasuredCpuTime(), 0.01);
		
//		Plotter p = new GnuplotPlotter(GnuplotPlotter.DFLT_PROPS);
//		p.display((Plot)run.getLastOutputValueOnly(AnalysisSemantics.SCD_OVERLAY));
//		p.display(((List<Plot>)run.getLastOutputValueOnly(AnalysisSemantics.SCDS)).get(0));
//		p.display(((List<Plot>)run.getLastOutputValueOnly(AnalysisSemantics.SCDS)).get(1));
//		p.display((Plot)run.getLastOutputValueOnly(AnalysisSemantics.SCATTER));
		
		System.out.println(run.getLastOutput());
		
		assertTrue(run.getLastOutputValueOnly(KWayComparison.FRIEDMAN_P) != null);
		assertTrue(run.getLastOutputValueOnly(AnalysisSemantics.SCD_OVERLAY) != null);
		
		// the first 3 are indistinguishable
		assertEquals(Arrays.asList(true, true, true, false, false), run.getLastOutputValueOnly(KWayComparison.WILCOXON_WINNERS));
	}
}
