package ca.ubc.cs.beta.smac;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Random;
import java.util.Set;

import org.apache.commons.io.FileUtils;
import org.junit.Before;
import org.junit.Test;

import ca.ubc.cs.beta.hal.algorithms.ParameterizedAlgorithm;
import ca.ubc.cs.beta.hal.algorithms.parameters.Domain;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSetting;
import ca.ubc.cs.beta.hal.algorithms.parameters.ParameterSettingBuilder;
import ca.ubc.cs.beta.hal.algorithms.parameters.SampleableDomain;
import ca.ubc.cs.beta.hal.environments.AlgorithmRun;
import ca.ubc.cs.beta.hal.problems.FileProblemInstance;
import ca.ubc.cs.beta.hal.problems.InstanceList;
import ca.ubc.cs.beta.hal.problems.ProblemInstance;
import ca.ubc.cs.beta.hal.problems.metrics.PerformanceMetric;
import ca.ubc.cs.beta.hal.utils.Global;
import ca.ubc.cs.beta.smac.InstanceSeedCombo;
import ca.ubc.cs.beta.smac.RunHistory;

public class RunHistoryTest {
	private final String cp = Global.isWindows() ? "bin;lib/*" : "bin:lib/*";
	RunHistory runHistory;
	Random rand;
	ParameterizedAlgorithm targetAlg;
	InstanceList instances;
	PerformanceMetric metric;
	HashSet<ProblemInstance> instanceSet;
	double delta = 1e-08;
	ProblemInstance instance0;
	ProblemInstance instance1;
	ProblemInstance instance2;
	ProblemInstance instance3;
	ParameterSetting config;
	ParameterSetting config2;
	boolean firstTime = true;

	public void onceInit() throws IOException {
		if (firstTime){	
			String json = "{'command':'java', 'name':'target', 'tags':[['foop']], 'outputFormat':{'stdout':['foroutput $v$ $t$']}, 'outputs':{'v':{'domain':'Real()'}, 't':{'domain':'Real()'}}, 'name':'json1', 'deterministic':false, 'inputFormat':{'callstring':['-classpath " + cp + " ca.ubc.cs.beta.hal.utils.TimeWaster $time$ [$hitime$]']}, 'inputs':{'time':{'domain':'Real[0, 0.5]', 'default':0.25}, 'hitime':{'domain':'Real[0.5, 1]', 'scale':'LOG', 'default':0.75}}}";
			targetAlg = ParameterizedAlgorithm.fromSpec(json);
			rand = Global.getRandom(1234);
			metric = PerformanceMetric.PAR1;
			
			// Dummy instances.
			LinkedList<ProblemInstance> il = new LinkedList<ProblemInstance>();
			for (int i=0; i<4; i++) {
				File f = File.createTempFile("dummy", ".inst", new File("."));
				f.deleteOnExit();
				FileUtils.writeStringToFile(f, f.getPath());
				il.add(new FileProblemInstance(f));
			}
			instances = new InstanceList("dummylist", il);
			instance0 = instances.get(0);
			instance1 = instances.get(1);
			instance2 = instances.get(2);
			instance3 = instances.get(3);
			
			instanceSet = new HashSet<ProblemInstance>(instances);
			
			config = getRandomConfig();
			config2 = getRandomConfig();
		} 
		firstTime = false;
	}
	
	@Before 
	public void init() throws IOException {
		onceInit();
		runHistory = new RunHistory();
	}

	@Test
	public void testGetEmpiricalCost() {
		assertEquals(Double.MAX_VALUE, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
		
		//=== basic tests for 1st configuration
		runHistory.append(config, instance0, 1, 1);
		assertEquals(1, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
		runHistory.append(config, instance0, 2, 11);
		assertEquals(6, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
		runHistory.append(config, instance0.clone(), 3, 9);
		assertEquals(7, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);

		//=== 2nd configuration
		assertEquals(Double.MAX_VALUE, runHistory.getEmpiricalCost(config2, metric, instanceSet), delta);
		runHistory.append(config2, instance0, 1, 1);
		assertEquals(7, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
		assertEquals(1, runHistory.getEmpiricalCost(config2, metric, instanceSet), delta);
		assertEquals(7, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
		
		//=== 2nd instance
		runHistory.append(config, instance1, 1, 11);
		assertEquals(9, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
	}
	
	@Test
	public void testGetInstanceSeedCombosRan() {
		Set<InstanceSeedCombo> result = runHistory.getInstanceSeedCombosRan(config);
		assert(result.isEmpty());
		
		runHistory.append(config, instance0, 1, 1);
		runHistory.append(config, instance0, 2, 1);
		runHistory.append(config, instance0, 12412, 1);
		runHistory.append(config, instance2, 12412, 1);

		Set<InstanceSeedCombo> instanceSeedCombosRan = new HashSet<InstanceSeedCombo>();
		instanceSeedCombosRan.add(new InstanceSeedCombo(instance0, 1));
		instanceSeedCombosRan.add(new InstanceSeedCombo(instance0, 2));
		instanceSeedCombosRan.add(new InstanceSeedCombo(instance0, 12412));
		instanceSeedCombosRan.add(new InstanceSeedCombo(instance2, 12412));

		result = runHistory.getInstanceSeedCombosRan(config);
		assert(result.containsAll(instanceSeedCombosRan));
		assert(instanceSeedCombosRan.containsAll(result));
	}
	
	
	@Test
	public void testGetInstancesRan() {
		Set<ProblemInstance> result = runHistory.getInstancesRan(config);
		assert(result.isEmpty());
		
		runHistory.append(config, instance0, 1, 1);
		runHistory.append(config, instance0, 2, 1);
		runHistory.append(config, instance0, 12412, 1);
		runHistory.append(config, instance2, 12412, 1);

		Set<ProblemInstance> instancesRan = new HashSet<ProblemInstance>();
		instancesRan.add(instance0);
		instancesRan.add(instance2);

		result = runHistory.getInstancesRan(config);
		assert(result.containsAll(instancesRan));
		assert(instancesRan.containsAll(result));
	}
	
	@Test
	public void testGetRandomInstanceWithFewestRunsFor() {
		//=== Probabilistic test: out of 2000 samples, at least 450 for each of the 4 options. 
		int inst0Count = 0;
		int inst1Count = 0;
		int inst2Count = 0;
		int inst3Count = 0;
		for (int i = 0; i < 2000; i++){
			ProblemInstance inst = runHistory.getRandomInstanceWithFewestRunsFor(config, instances, rand);
			if (inst.equals(instance0)) inst0Count++;
			if (inst.equals(instance1)) inst1Count++;
			if (inst.equals(instance2)) inst2Count++;
			if (inst.equals(instance3))	inst3Count++;
			assert(instances.contains(inst));
		}
		assert(inst0Count>=450);
		assert(inst1Count>=450);
		assert(inst2Count>=450);
		assert(inst3Count>=450);
		
		//=== Probabilistic test: out of 1000 samples, at least 450 for each of the 2 options.  
		runHistory.append(config, instance0, 1, 1);
		runHistory.append(config, instance1, 2, 1);
		Set<ProblemInstance> inst2and3 = new HashSet<ProblemInstance>();
		inst2and3.add(instance2);
		inst2and3.add(instance3);
		for (int i = 0; i < 1000; i++){
			ProblemInstance inst = runHistory.getRandomInstanceWithFewestRunsFor(config, instances, rand);
			if (inst.equals(instance2)) inst2Count++;
			if (inst.equals(instance3))	inst3Count++;
			assert(inst2and3.contains(inst));
		}
		assert(inst2Count>=450);
		assert(inst3Count>=450);
		
		runHistory.append(config, instance2, 1, 11);
		for (int i = 0; i < 1000; i++){
			assertEquals(instance3, runHistory.getRandomInstanceWithFewestRunsFor(config, instances, rand));
		}
	}
	
	@Test
	public void testGetTotalNumRunsOfConfig() {
		runHistory.append(config, instance0, 1, 1);
		runHistory.append(config, instance0, 2, 1);
		runHistory.append(config, instance0, 12412, 1);
		runHistory.append(config, instance2, 12412, 1);

		assertEquals( runHistory.getTotalNumRunsOfConfig(config), 4);
		assertEquals( runHistory.getTotalNumRunsOfConfig(config2), 0);
		
		runHistory.append(config2, instance0, 1, 1);
		runHistory.append(config2, instance0, 2, 1);
		runHistory.append(config2, instance0, 12412, 1);
		runHistory.append(config2, instance2, 12412, 1);
		
		assertEquals( runHistory.getTotalNumRunsOfConfig(config), 4);
		assertEquals( runHistory.getTotalNumRunsOfConfig(config2), 4);

		for (int i = 0; i < 1000; i++){
			runHistory.append(config, instance0.clone(), i+4, 10);
		}
		assertEquals( runHistory.getTotalNumRunsOfConfig(config), 1004);
		assertEquals( runHistory.getTotalNumRunsOfConfig(config2), 4);
	}	
	
	@Test
	public void testDuplicates() {
		runHistory.append(config, instance0, 1, 1);
		assertEquals( runHistory.getTotalNumRunsOfConfig(config), 1);
		
		try {
			runHistory.append(config, instance0, 1, 2);
			assert(false);
		} catch (RuntimeException re) {
			// we *want* this exception
		}
		
		//=== Not counting duplicates in any of the methods:
		assertEquals( runHistory.getTotalNumRunsOfConfig(config), 1);
		
		Set<ProblemInstance> instancesRan = new HashSet<ProblemInstance>();
		instancesRan.add(instance0);
		Set<ProblemInstance> result = runHistory.getInstancesRan(config);
		assertEquals(result, instancesRan);

		Set<InstanceSeedCombo> instanceCombosRan = new HashSet<InstanceSeedCombo>();
		instanceCombosRan.add(new InstanceSeedCombo(instance0, 1));
		Set<InstanceSeedCombo> result2 = runHistory.getInstanceSeedCombosRan(config);
		assertEquals(result2, instanceCombosRan);
		
		assertEquals(1, runHistory.getEmpiricalCost(config, metric, instanceSet), delta);
	}
	
	@Test
	public void testRobustnessAgainstManipulationOfOutputs() {
		runHistory.append(config, instance0, 1, 1);
		
		Set<ProblemInstance> origResult = new HashSet<ProblemInstance>(runHistory.getInstancesRan(config));
		Set<ProblemInstance> result = runHistory.getInstancesRan(config);
		assertEquals(result, origResult);
		result.clear();
		result = runHistory.getInstancesRan(config);
		assertEquals(result, origResult);
		
		Set<InstanceSeedCombo> origResultI = new HashSet<InstanceSeedCombo>(runHistory.getInstanceSeedCombosRan(config));
		Set<InstanceSeedCombo> resultI = runHistory.getInstanceSeedCombosRan(config);
		assertEquals(resultI, origResultI);

		resultI.clear();
		resultI = runHistory.getInstanceSeedCombosRan(config);
		assertEquals(resultI, origResultI);
	}	
	
	/* Copied from RunHistory. Copying once for the UnitTest is not horrible... */
	protected ParameterSetting getRandomConfig() {	
		Set<String> parameterNames = targetAlg.getConfiguration().keySet();
		ParameterSettingBuilder randomConfig = new ParameterSettingBuilder();
		
		// get a new value for each key
		for ( String key : parameterNames ) {
			Domain d = targetAlg.getConfigurationDomain(key);
			if ( d instanceof SampleableDomain<?> ) {
				randomConfig.put( key, ((SampleableDomain<?>) d).getSample( rand ));
			}
		}
		
		//=== Check if parameters are active, otherwise remove them.
		// TODO: sample in topological order to begin with.
		for ( String key: parameterNames ) {
			if ( !targetAlg.isActiveInConfiguration(key, randomConfig) ) {
				randomConfig.remove(key);
			}
		}
		
		return randomConfig.build();
	}

}
